from pathlib import Path
from typing import List
from clang.cindex import Index as CIndex, Cursor, TranslationUnit
import os
import hashlib
import pickle

from excal.astNode import AstNode, Location, Token


class CWalker:
    """Create AST and translate it to astNodes"""
    def __init__(self, c_file: Path, clang_args: List[str], cache: bool, baseDir: Path) -> None:
        self.anonymous_counter = 0
        self.path: Path = c_file
        self.args = clang_args
        # use clang to parse C file
        self.index: CIndex = CIndex.create()
        self.use_cache = cache
        self.baseDir = baseDir
        self.parsed_unit: TranslationUnit = None
        self.root_node: Cursor = None

        if self.use_cache:
            self.excalDir = self.baseDir / ".excal"
            if not os.path.isdir(self.excalDir):
                os.mkdir(self.excalDir)
            file = open(self.path)
            fileC = file.read()
            hash = hashlib.sha256(fileC.encode())
            file.close()
            self.hash = hash.hexdigest()

            if not os.path.isfile(self.excalDir / self.hash):
                self.parsed_unit = self.index.parse(self.path, args=self.args)
                self.root_node = self.parsed_unit.cursor
                # TL.save(self.excalDir / hash.hexdigest())
        else:
            self.parsed_unit = self.index.parse(self.path, args=self.args)
            self.root_node = self.parsed_unit.cursor

        self.ast: AstNode

    def walkRec(self, node: Cursor, indent: str, ast: AstNode) -> None:
        indent += '  '
        for child_node in node.get_children():
            try:
                if not self.path.samefile(child_node.location.file.name):
                    continue
            except Exception:
                continue
            ast_child = AstNode(str(child_node.kind), child_node.location.file.name,
                                child_node.location.line, child_node.location.column,
                                child_node.extent.end.line, child_node.extent.end.column,
                                str(child_node.spelling), ast.indent_level + 1,
                                str(child_node.type.spelling), ast,
                                [Token(str(x.kind), x.spelling, Location(x.location.line, x.location.column)) for x in child_node.get_tokens()])

            ast.add_child(ast_child)
            self.walkRec(child_node, indent, ast_child)

    def walk(self) -> AstNode:
        self.excalDir = self.baseDir / ".excal"

        if self.use_cache and os.path.isfile(self.excalDir / self.hash):
            cFile = open(self.excalDir / self.hash, "rb")
            self.ast = pickle.load(cFile)
        else:
            self.ast = AstNode(str(self.root_node.kind), "", 0, 0, self.root_node.extent.end.line, self.root_node.extent.end.column, str(self.root_node.spelling), 0, "", None, [Token(str(x.kind), x.spelling, Location(x.location.line, x.location.column)) for x in self.root_node.get_tokens()])
            self.walkRec(self.root_node, '', self.ast)

        if self.use_cache and not os.path.isfile(self.excalDir / self.hash):
            with open(self.excalDir / self.hash, 'wb') as fh:
                pickle.dump(self.ast, fh)
        return self.ast



from dataclasses import dataclass
from enum import Enum
from excal.astNode import Location

class OffenceType(Enum):
    CONVENTION = 0
    CODE_SMELL = 1
    VULNERABILITY = 2
    BUG = 3

class Severity(Enum):
    CONVENTION = 0
    INFO = 1
    MINOR = 2
    MAJOR = 3
    CRITICAL = 4
    BLOCKER = 5


@dataclass
class Offence():
    file: str
    start_location: Location
    message: str
    id: str
    end_location: Location = Location(0, 0)
    severity: Severity = Severity.CONVENTION
    type: OffenceType = OffenceType.CONVENTION



"""EXCAL - Extendable Clang AST based Linter."""
import os
from pathlib import Path
import glob
from typing import List
import configparser
from argparse import ArgumentParser

from excal.analyzer import Analyzer
from excal.pluginManager import PluginManager
from excal.output import Output


def findConfig(path: str, find_config: bool) -> str:

    if os.path.exists(f"{path}setup.cfg"):
        return path
    if not find_config:
        return ""

    walkPath = Path(path)
    # parent function will return previous path if no parent exists so this checks if root node exists.
    while walkPath.parent != path:
        walkPath = walkPath.parent
        if os.path.exists(f"{str(walkPath)}/setup.cfg"):
            return str(walkPath) + "/"
    return ""


def readConfig(args):
    if len(args.files) != 1:
        return args

    path: str = args.files[0]

    if not os.path.isdir(path) and not args.find_config:
        return args
    if not path.endswith('/') and not path.endswith('\\'):
        path += '/'
    if not os.path.exists(f"{path}setup.cfg") and not args.find_config:
        return args

    config = configparser.ConfigParser()
    configPath = findConfig(path, args.find_config)
    if configPath == "":
        return args

    config.read(f"{configPath}setup.cfg")

    if 'EXCAL' not in config.sections():
        return args

    cfg = config['EXCAL']
    for key in cfg:
        if key == "force_cpp":
            args.force_cpp = cfg[key] == 'True'
        if key == "includes":
            args.include.extend([l.strip() for l in cfg[key].split('\n')])
        if key == "exclude_files" and not args.find_config:
            args.exclude_files.extend(
                [f'{path}{line.strip()}'
                 for line in cfg[key].split('\n') if line != ''])
        # files überschreiben das directory. Wenn dieses auch gescannt werden soll, muss es auch eingefügt werden.
        if key == "parse_files" and not args.find_config:
            args.files = [f'{path}{line.strip()}'
                          for line in cfg[key].split('\n') if line != '']

    return args


def main() -> None:
    parser = ArgumentParser()
    parser.add_argument("--args", "-a", nargs="*", required=False,
                        default=[], help="Additional clang arguments")
    parser.add_argument("--argfile", type=str, default=None, required=False,
                        help="File containing clang arguments")
    parser.add_argument("--include", "-i", nargs="+", required=False, default=[],
                        help="include directories")
    parser.add_argument("--files", "-f", nargs="+", required=True,
                        help="File to translate")
    parser.add_argument("--exclude-files", "-e", nargs="+", required=False, default=[],
                        help="files within files directory, that should be ignored")
    parser.add_argument("--print-tree", "-p", nargs='?', required=False,
                        const=True, default=False, help="print AST of input files")
    parser.add_argument("--force-cpp", "-cpp", nargs='?', required=False,
                        const=True, default=False,
                        help="forces files to be analyzed as C++ files.")
    parser.add_argument("--use-chache", "-c", nargs='?', required=False,
                        const=True, default=False,
                        help="Will safe AST files in .excal Folder and try to access said files to speedup redundant loading.")
    parser.add_argument("--extensions", "-ext", nargs="+", required=False,
                        default=[], help="file extensions to parse.")
    parser.add_argument("--print-token-tree", "-ptt", nargs='?', required=False,
                        const=True, default=False, help="print AST plus corresponding Tokens")
    parser.add_argument("--output-style", "-out", choices=['stdErr', 'json', 'sonarqube'], required=False,
                        default='stdErr', help="file extensions to parse.")
    parser.add_argument("--find-config", "-fcfg", nargs='?', required=False,
                        const=True, default=False,
                        help="Will try and find a setup.cfg file in parents directories of input-file. Will ignore\
                        parse_files and exclude_files options in .cfg file, as it may be used for single files.")

    args = parser.parse_args()

    baseDir = Path(args.files[0]) if os.path.isdir(args.files[0]) else Path(args.files[0]).parent()
    args = readConfig(args)
    print(baseDir)
    clang_args = args.args

    if args.argfile is not None:
        with open(args.argfile, "r") as fl:
            lines = [ln.strip() for ln in fl.readlines()]
            clang_args.extend(lines)

    if args.force_cpp:
        clang_args.extend(['-x', 'c++'])

    FILE_EXTENSIONS: List[str] = []
    if args.extensions != []:
        FILE_EXTENSIONS = args.extensions
    else:
        FILE_EXTENSIONS = ["c", "h", "cpp", "hpp"]

    exfiles = set()
    for ex_file in args.exclude_files:
        if not os.path.isfile(ex_file):
            for ext in FILE_EXTENSIONS:
                exfiles |= set(glob.glob(f"{ex_file}/**/*.{ext}", recursive=True))
        else:
            exfiles.add(ex_file)

    analyze_files: List[str] = []
    for in_file in args.files:
        if os.path.isdir(in_file):
            for ext in FILE_EXTENSIONS:
                new_files = glob.glob(f"{in_file}/**/*.{ext}", recursive=True)
                analyze_files = analyze_files + [f for f in new_files if f not in exfiles]
        else:
            analyze_files = analyze_files + [in_file]

    pm: PluginManager = PluginManager()
    pm.loadPlugins()
    out: Output = Output()
    analyzer = Analyzer(args.include, clang_args, pm, out, args.use_chache, baseDir)

    printState = 0
    if(args.print_tree):
        printState = 1
    if(args.print_token_tree):
        printState = 2

    for in_file in analyze_files:
        analyzer.analyze(in_file, printState)

    if args.output_style == 'stdErr':
        out.printError()
    if args.output_style == "json":
        out.printJson()
    if args.output_style == "sonarqube":
        out.printSonarqube()


if __name__ == "__main__":
    main()



import sys
import json
from typing import List
from excal.offence import Offence


class Output():
    """Output class. Will be extended by more output options."""
    def __init__(self):
        self.exeptions: List[Offence] = []
        return

    def addExeptioins(self, ex: List[Offence]):
        self.exeptions.extend(ex)

    def printError(self):
        for e in self.exeptions:
            print(f'{e.file}: {e.start_location.line}:{e.start_location.col} : {e.id} {e.message}', file=sys.stderr)

    def printJson(self):
        out = json.loads('{"issues":[]}')
        for o in self.exeptions:
            a = o.__dict__
            a["start_location"] = o.start_location.__dict__
            a["end_location"] = o.end_location.__dict__
            a["severity"] = o.severity.name
            a["type"] = o.type.name
            out["issues"].append(a)

        print(json.dumps(out, indent=2))

    def printSonarqube(self):
        out = json.loads('{"issues":[]}')
        for o in self.exeptions:
            a = {
                "engineId": "EXCAL",
                "ruleId": o.id,
                "severity": "INFO" if o.severity == 0 else o.severity.name,
                "type": "CODE_SMELL" if o.type == 0 else o.type.name,
            }
            a["primaryLocation"] = {
                "message": o.message,
                "filePath": o.file
            }
            a["primaryLocation"]["textRange"] = {
                "startLine": o.start_location.line,
                "endLine": o.end_location.line,
                "startColumn": o.start_location.col,
                "endColumn": o.end_location.col
            }

            a["start_location"] = o.start_location.__dict__
            a["end_location"] = o.end_location.__dict__
            a["severity"] = o.severity.name
            a["type"] = o.type.name
            out["issues"].append(a)
        print(json.dumps(out, indent=2))



from typing import Callable, Dict, List
from typing import Generator
from clang.cindex import CursorKind

from excal.astNode import AstNode
from excal.pluginManager import PluginManager
from excal.offence import Offence


class CXXAstNodeVisitor():
    """This class will run through the AST, and call all Plugins on each node, they wish to be called at."""
    def __init__(self, ast, pm: PluginManager):
        self.ast = ast
        self.visitors: List[NodeVisitor] = pm.getPluginList()
        self.exceptions: List[Offence] = []

    def visitRec(self, node: AstNode) -> None:
        for visitor in self.visitors:
            exepts = visitor.visit(node)
            if exepts is not None:
                self.exceptions.extend([e for e in exepts])
        for node_child in node.children:
            self.visitRec(node_child)

    def visit(self) -> None:
        self.visitRec(self.ast)
        for visitor in self.visitors:
            exepts = visitor.final_call()
            if exepts is not None:
                self.exceptions.extend([e for e in exepts])

    def getExceptions(self) -> List[Offence]:
        return self.exceptions


class NodeVisitor():
    """Visitor class. Plugins will inherit from it."""
    def __init__(self):
        self._jumpTable: Dict[str, Callable[AstNode, Generator[Offence, None, None]]] = {}
        self._jumpTable[str(CursorKind.TRANSLATION_UNIT)] = self.visit_translation_unit
        self._jumpTable[str(CursorKind.UNEXPOSED_DECL)] = self.visit_unexposed_decl
        self._jumpTable[str(CursorKind.STRUCT_DECL)] = self.visit_struct_decl
        self._jumpTable[str(CursorKind.UNION_DECL)] = self.visit_union_decl
        self._jumpTable[str(CursorKind.CLASS_DECL)] = self.visit_class_decl
        self._jumpTable[str(CursorKind.ENUM_DECL)] = self.visit_enum_decl
        self._jumpTable[str(CursorKind.FIELD_DECL)] = self.visit_field_decl
        self._jumpTable[str(CursorKind.ENUM_CONSTANT_DECL)] = self.visit_enum_constant_decl
        self._jumpTable[str(CursorKind.FUNCTION_DECL)] = self.visit_function_decl
        self._jumpTable[str(CursorKind.VAR_DECL)] = self.visit_var_decl
        self._jumpTable[str(CursorKind.PARM_DECL)] = self.visit_param_decl
        self._jumpTable[str(CursorKind.OBJC_INTERFACE_DECL)] = self.visit_objc_interface_decl
        # TODO objective C parameter vowegelassen
        self._jumpTable[str(CursorKind.TYPEDEF_DECL)] = self.visit_typedef_decl
        self._jumpTable[str(CursorKind.CXX_METHOD)] = self.visit_cxx_method
        self._jumpTable[str(CursorKind.NAMESPACE)] = self.visit_namespace
        self._jumpTable[str(CursorKind.CXX_BASE_SPECIFIER)] = self.visit_cxx_base_specifier
        # from here autogenerated, may contain errors

        self._jumpTable[str(CursorKind.LINKAGE_SPEC)] = self.visit_linkage_spec
        self._jumpTable[str(CursorKind.CONSTRUCTOR)] = self.visit_constructor
        self._jumpTable[str(CursorKind.DESTRUCTOR)] = self.visit_destructor
        self._jumpTable[str(CursorKind.CONVERSION_FUNCTION)] = self.visit_conversion_function
        self._jumpTable[str(CursorKind.TEMPLATE_TYPE_PARAMETER)] = self.visit_template_type_parameter
        self._jumpTable[str(CursorKind.TEMPLATE_NON_TYPE_PARAMETER)] = self.visit_template_non_type_parameter
        self._jumpTable[str(CursorKind.TEMPLATE_TEMPLATE_PARAMETER)] = self.visit_template_template_parameter
        self._jumpTable[str(CursorKind.FUNCTION_TEMPLATE)] = self.visit_function_template
        self._jumpTable[str(CursorKind.CLASS_TEMPLATE)] = self.visit_class_template
        self._jumpTable[str(CursorKind.CLASS_TEMPLATE_PARTIAL_SPECIALIZATION)] = \
            self.visit_class_template_partial_specialization
        self._jumpTable[str(CursorKind.NAMESPACE_ALIAS)] = self.visit_namespace_alias
        self._jumpTable[str(CursorKind.USING_DIRECTIVE)] = self.visit_using_directive
        self._jumpTable[str(CursorKind.USING_DECLARATION)] = self.visit_using_declaration
        self._jumpTable[str(CursorKind.TYPE_ALIAS_DECL)] = self.visit_type_alias_decl
        self._jumpTable[str(CursorKind.OBJC_SYNTHESIZE_DECL)] = self.visit_objc_synthesize_decl
        self._jumpTable[str(CursorKind.OBJC_DYNAMIC_DECL)] = self.visit_objc_dynamic_decl
        self._jumpTable[str(CursorKind.CXX_ACCESS_SPEC_DECL)] = self.visit_cxx_access_spec_decl
        self._jumpTable[str(CursorKind.OBJC_SUPER_CLASS_REF)] = self.visit_objc_super_class_ref
        self._jumpTable[str(CursorKind.OBJC_PROTOCOL_REF)] = self.visit_objc_protocol_ref
        self._jumpTable[str(CursorKind.OBJC_CLASS_REF)] = self.visit_objc_class_ref
        self._jumpTable[str(CursorKind.TYPE_REF)] = self.visit_type_ref
        self._jumpTable[str(CursorKind.TEMPLATE_REF)] = self.visit_template_ref
        self._jumpTable[str(CursorKind.NAMESPACE_REF)] = self.visit_namespace_ref
        self._jumpTable[str(CursorKind.MEMBER_REF)] = self.visit_member_ref
        self._jumpTable[str(CursorKind.LABEL_REF)] = self.visit_label_ref
        self._jumpTable[str(CursorKind.OVERLOADED_DECL_REF)] = self.visit_overloaded_decl_ref
        self._jumpTable[str(CursorKind.VARIABLE_REF)] = self.visit_variable_ref
        self._jumpTable[str(CursorKind.INVALID_FILE)] = self.visit_invalid_file
        self._jumpTable[str(CursorKind.NO_DECL_FOUND)] = self.visit_no_decl_found
        self._jumpTable[str(CursorKind.NOT_IMPLEMENTED)] = self.visit_not_implemented
        self._jumpTable[str(CursorKind.INVALID_CODE)] = self.visit_invalid_code
        self._jumpTable[str(CursorKind.UNEXPOSED_EXPR)] = self.visit_unexposed_expr
        self._jumpTable[str(CursorKind.DECL_REF_EXPR)] = self.visit_decl_ref_expr
        self._jumpTable[str(CursorKind.MEMBER_REF_EXPR)] = self.visit_member_ref_expr
        self._jumpTable[str(CursorKind.CALL_EXPR)] = self.visit_call_expr
        self._jumpTable[str(CursorKind.OBJC_MESSAGE_EXPR)] = self.visit_objc_message_expr
        self._jumpTable[str(CursorKind.BLOCK_EXPR)] = self.visit_block_expr
        self._jumpTable[str(CursorKind.INTEGER_LITERAL)] = self.visit_integer_literal
        self._jumpTable[str(CursorKind.FLOATING_LITERAL)] = self.visit_floating_literal
        self._jumpTable[str(CursorKind.IMAGINARY_LITERAL)] = self.visit_imaginary_literal
        self._jumpTable[str(CursorKind.STRING_LITERAL)] = self.visit_string_literal
        self._jumpTable[str(CursorKind.CHARACTER_LITERAL)] = self.visit_character_literal
        self._jumpTable[str(CursorKind.PAREN_EXPR)] = self.visit_paren_expr
        self._jumpTable[str(CursorKind.UNARY_OPERATOR)] = self.visit_unary_operator
        self._jumpTable[str(CursorKind.ARRAY_SUBSCRIPT_EXPR)] = self.visit_array_subscript_expr
        self._jumpTable[str(CursorKind.BINARY_OPERATOR)] = self.visit_binary_operator
        self._jumpTable[str(CursorKind.COMPOUND_ASSIGNMENT_OPERATOR)] = self.visit_compound_assignment_operator
        self._jumpTable[str(CursorKind.CONDITIONAL_OPERATOR)] = self.visit_conditional_operator
        self._jumpTable[str(CursorKind.CSTYLE_CAST_EXPR)] = self.visit_cstyle_cast_expr
        self._jumpTable[str(CursorKind.COMPOUND_LITERAL_EXPR)] = self.visit_compound_literal_expr
        self._jumpTable[str(CursorKind.INIT_LIST_EXPR)] = self.visit_init_list_expr
        self._jumpTable[str(CursorKind.ADDR_LABEL_EXPR)] = self.visit_addr_label_expr
        self._jumpTable[str(CursorKind.StmtExpr)] = self.visit_stmtexpr
        self._jumpTable[str(CursorKind.GENERIC_SELECTION_EXPR)] = self.visit_generic_selection_expr
        self._jumpTable[str(CursorKind.GNU_NULL_EXPR)] = self.visit_gnu_null_expr
        self._jumpTable[str(CursorKind.CXX_STATIC_CAST_EXPR)] = self.visit_cxx_static_cast_expr
        self._jumpTable[str(CursorKind.CXX_DYNAMIC_CAST_EXPR)] = self.visit_cxx_dynamic_cast_expr
        self._jumpTable[str(CursorKind.CXX_REINTERPRET_CAST_EXPR)] = self.visit_cxx_reinterpret_cast_expr
        self._jumpTable[str(CursorKind.CXX_CONST_CAST_EXPR)] = self.visit_cxx_const_cast_expr
        self._jumpTable[str(CursorKind.CXX_FUNCTIONAL_CAST_EXPR)] = self.visit_cxx_functional_cast_expr
        self._jumpTable[str(CursorKind.CXX_TYPEID_EXPR)] = self.visit_cxx_typeid_expr
        self._jumpTable[str(CursorKind.CXX_BOOL_LITERAL_EXPR)] = self.visit_cxx_bool_literal_expr
        self._jumpTable[str(CursorKind.CXX_NULL_PTR_LITERAL_EXPR)] = self.visit_cxx_null_ptr_literal_expr
        self._jumpTable[str(CursorKind.CXX_THIS_EXPR)] = self.visit_cxx_this_expr
        self._jumpTable[str(CursorKind.CXX_THROW_EXPR)] = self.visit_cxx_throw_expr
        self._jumpTable[str(CursorKind.CXX_NEW_EXPR)] = self.visit_cxx_new_expr
        self._jumpTable[str(CursorKind.CXX_DELETE_EXPR)] = self.visit_cxx_delete_expr
        self._jumpTable[str(CursorKind.CXX_UNARY_EXPR)] = self.visit_cxx_unary_expr
        self._jumpTable[str(CursorKind.OBJC_STRING_LITERAL)] = self.visit_objc_string_literal
        self._jumpTable[str(CursorKind.OBJC_ENCODE_EXPR)] = self.visit_objc_encode_expr
        self._jumpTable[str(CursorKind.OBJC_SELECTOR_EXPR)] = self.visit_objc_selector_expr
        self._jumpTable[str(CursorKind.OBJC_PROTOCOL_EXPR)] = self.visit_objc_protocol_expr
        self._jumpTable[str(CursorKind.OBJC_BRIDGE_CAST_EXPR)] = self.visit_objc_bridge_cast_expr
        self._jumpTable[str(CursorKind.PACK_EXPANSION_EXPR)] = self.visit_pack_expansion_expr
        self._jumpTable[str(CursorKind.SIZE_OF_PACK_EXPR)] = self.visit_size_of_pack_expr
        self._jumpTable[str(CursorKind.LAMBDA_EXPR)] = self.visit_lambda_expr
        self._jumpTable[str(CursorKind.OBJ_BOOL_LITERAL_EXPR)] = self.visit_obj_bool_literal_expr
        self._jumpTable[str(CursorKind.OBJ_SELF_EXPR)] = self.visit_obj_self_expr
        self._jumpTable[str(CursorKind.OMP_ARRAY_SECTION_EXPR)] = self.visit_omp_array_section_expr
        self._jumpTable[str(CursorKind.OBJC_AVAILABILITY_CHECK_EXPR)] = self.visit_objc_availability_check_expr
        self._jumpTable[str(CursorKind.UNEXPOSED_STMT)] = self.visit_unexposed_stmt
        self._jumpTable[str(CursorKind.LABEL_STMT)] = self.visit_label_stmt
        self._jumpTable[str(CursorKind.COMPOUND_STMT)] = self.visit_compound_stmt
        self._jumpTable[str(CursorKind.CASE_STMT)] = self.visit_case_stmt
        self._jumpTable[str(CursorKind.DEFAULT_STMT)] = self.visit_default_stmt
        self._jumpTable[str(CursorKind.IF_STMT)] = self.visit_if_stmt
        self._jumpTable[str(CursorKind.SWITCH_STMT)] = self.visit_switch_stmt
        self._jumpTable[str(CursorKind.WHILE_STMT)] = self.visit_while_stmt
        self._jumpTable[str(CursorKind.DO_STMT)] = self.visit_do_stmt
        self._jumpTable[str(CursorKind.FOR_STMT)] = self.visit_for_stmt
        self._jumpTable[str(CursorKind.GOTO_STMT)] = self.visit_goto_stmt
        self._jumpTable[str(CursorKind.INDIRECT_GOTO_STMT)] = self.visit_indirect_goto_stmt
        self._jumpTable[str(CursorKind.CONTINUE_STMT)] = self.visit_continue_stmt
        self._jumpTable[str(CursorKind.BREAK_STMT)] = self.visit_break_stmt
        self._jumpTable[str(CursorKind.RETURN_STMT)] = self.visit_return_stmt
        self._jumpTable[str(CursorKind.ASM_STMT)] = self.visit_asm_stmt
        self._jumpTable[str(CursorKind.OBJC_AT_TRY_STMT)] = self.visit_objc_at_try_stmt
        self._jumpTable[str(CursorKind.OBJC_AT_CATCH_STMT)] = self.visit_objc_at_catch_stmt
        self._jumpTable[str(CursorKind.OBJC_AT_FINALLY_STMT)] = self.visit_objc_at_finally_stmt
        self._jumpTable[str(CursorKind.OBJC_AT_THROW_STMT)] = self.visit_objc_at_throw_stmt
        self._jumpTable[str(CursorKind.OBJC_AT_SYNCHRONIZED_STMT)] = self.visit_objc_at_synchronized_stmt
        self._jumpTable[str(CursorKind.OBJC_AUTORELEASE_POOL_STMT)] = self.visit_objc_autorelease_pool_stmt
        self._jumpTable[str(CursorKind.OBJC_FOR_COLLECTION_STMT)] = self.visit_objc_for_collection_stmt
        self._jumpTable[str(CursorKind.CXX_CATCH_STMT)] = self.visit_cxx_catch_stmt
        self._jumpTable[str(CursorKind.CXX_TRY_STMT)] = self.visit_cxx_try_stmt
        self._jumpTable[str(CursorKind.CXX_FOR_RANGE_STMT)] = self.visit_cxx_for_range_stmt
        self._jumpTable[str(CursorKind.SEH_TRY_STMT)] = self.visit_seh_try_stmt
        self._jumpTable[str(CursorKind.SEH_EXCEPT_STMT)] = self.visit_seh_except_stmt
        self._jumpTable[str(CursorKind.SEH_FINALLY_STMT)] = self.visit_seh_finally_stmt
        self._jumpTable[str(CursorKind.MS_ASM_STMT)] = self.visit_ms_asm_stmt
        self._jumpTable[str(CursorKind.NULL_STMT)] = self.visit_null_stmt
        self._jumpTable[str(CursorKind.DECL_STMT)] = self.visit_decl_stmt
        self._jumpTable[str(CursorKind.OMP_PARALLEL_DIRECTIVE)] = self.visit_omp_parallel_directive
        self._jumpTable[str(CursorKind.OMP_SIMD_DIRECTIVE)] = self.visit_omp_simd_directive
        self._jumpTable[str(CursorKind.OMP_FOR_DIRECTIVE)] = self.visit_omp_for_directive
        self._jumpTable[str(CursorKind.OMP_SECTIONS_DIRECTIVE)] = self.visit_omp_sections_directive
        self._jumpTable[str(CursorKind.OMP_SECTION_DIRECTIVE)] = self.visit_omp_section_directive
        self._jumpTable[str(CursorKind.OMP_SINGLE_DIRECTIVE)] = self.visit_omp_single_directive
        self._jumpTable[str(CursorKind.OMP_PARALLEL_FOR_DIRECTIVE)] = self.visit_omp_parallel_for_directive
        self._jumpTable[str(CursorKind.OMP_PARALLEL_SECTIONS_DIRECTIVE)] = self.visit_omp_parallel_sections_directive
        self._jumpTable[str(CursorKind.OMP_TASK_DIRECTIVE)] = self.visit_omp_task_directive
        self._jumpTable[str(CursorKind.OMP_MASTER_DIRECTIVE)] = self.visit_omp_master_directive
        self._jumpTable[str(CursorKind.OMP_CRITICAL_DIRECTIVE)] = self.visit_omp_critical_directive
        self._jumpTable[str(CursorKind.OMP_TASKYIELD_DIRECTIVE)] = self.visit_omp_taskyield_directive
        self._jumpTable[str(CursorKind.OMP_BARRIER_DIRECTIVE)] = self.visit_omp_barrier_directive
        self._jumpTable[str(CursorKind.OMP_TASKWAIT_DIRECTIVE)] = self.visit_omp_taskwait_directive
        self._jumpTable[str(CursorKind.OMP_FLUSH_DIRECTIVE)] = self.visit_omp_flush_directive
        self._jumpTable[str(CursorKind.SEH_LEAVE_STMT)] = self.visit_seh_leave_stmt
        self._jumpTable[str(CursorKind.OMP_ORDERED_DIRECTIVE)] = self.visit_omp_ordered_directive
        self._jumpTable[str(CursorKind.OMP_ATOMIC_DIRECTIVE)] = self.visit_omp_atomic_directive
        self._jumpTable[str(CursorKind.OMP_FOR_SIMD_DIRECTIVE)] = self.visit_omp_for_simd_directive
        self._jumpTable[str(CursorKind.OMP_PARALLELFORSIMD_DIRECTIVE)] = self.visit_omp_parallelforsimd_directive
        self._jumpTable[str(CursorKind.OMP_TARGET_DIRECTIVE)] = self.visit_omp_target_directive
        self._jumpTable[str(CursorKind.OMP_TEAMS_DIRECTIVE)] = self.visit_omp_teams_directive
        self._jumpTable[str(CursorKind.OMP_TASKGROUP_DIRECTIVE)] = self.visit_omp_taskgroup_directive
        self._jumpTable[str(CursorKind.OMP_CANCELLATION_POINT_DIRECTIVE)] = self.visit_omp_cancellation_point_directive
        self._jumpTable[str(CursorKind.OMP_CANCEL_DIRECTIVE)] = self.visit_omp_cancel_directive
        self._jumpTable[str(CursorKind.OMP_TARGET_DATA_DIRECTIVE)] = self.visit_omp_target_data_directive
        self._jumpTable[str(CursorKind.OMP_TASK_LOOP_DIRECTIVE)] = self.visit_omp_task_loop_directive
        self._jumpTable[str(CursorKind.OMP_TASK_LOOP_SIMD_DIRECTIVE)] = self.visit_omp_task_loop_simd_directive
        self._jumpTable[str(CursorKind.OMP_DISTRIBUTE_DIRECTIVE)] = self.visit_omp_distribute_directive
        self._jumpTable[str(CursorKind.OMP_TARGET_ENTER_DATA_DIRECTIVE)] = self.visit_omp_target_enter_data_directive
        self._jumpTable[str(CursorKind.OMP_TARGET_EXIT_DATA_DIRECTIVE)] = self.visit_omp_target_exit_data_directive
        self._jumpTable[str(CursorKind.OMP_TARGET_PARALLEL_DIRECTIVE)] = self.visit_omp_target_parallel_directive
        self._jumpTable[str(CursorKind.OMP_TARGET_PARALLELFOR_DIRECTIVE)] = self.visit_omp_target_parallelfor_directive
        self._jumpTable[str(CursorKind.OMP_TARGET_UPDATE_DIRECTIVE)] = self.visit_omp_target_update_directive
        self._jumpTable[str(CursorKind.OMP_DISTRIBUTE_PARALLELFOR_DIRECTIVE)] = \
            self.visit_omp_distribute_parallelfor_directive
        self._jumpTable[str(CursorKind.OMP_DISTRIBUTE_PARALLEL_FOR_SIMD_DIRECTIVE)] = \
            self.visit_omp_distribute_parallel_for_simd_directive
        self._jumpTable[str(CursorKind.OMP_DISTRIBUTE_SIMD_DIRECTIVE)] = self.visit_omp_distribute_simd_directive
        self._jumpTable[str(CursorKind.OMP_TARGET_PARALLEL_FOR_SIMD_DIRECTIVE)] = \
            self.visit_omp_target_parallel_for_simd_directive
        self._jumpTable[str(CursorKind.OMP_TARGET_SIMD_DIRECTIVE)] = self.visit_omp_target_simd_directive
        self._jumpTable[str(CursorKind.OMP_TEAMS_DISTRIBUTE_DIRECTIVE)] = self.visit_omp_teams_distribute_directive
        self._jumpTable[str(CursorKind.UNEXPOSED_ATTR)] = self.visit_unexposed_attr
        self._jumpTable[str(CursorKind.IB_ACTION_ATTR)] = self.visit_ib_action_attr
        self._jumpTable[str(CursorKind.IB_OUTLET_ATTR)] = self.visit_ib_outlet_attr
        self._jumpTable[str(CursorKind.IB_OUTLET_COLLECTION_ATTR)] = self.visit_ib_outlet_collection_attr
        self._jumpTable[str(CursorKind.CXX_FINAL_ATTR)] = self.visit_cxx_final_attr
        self._jumpTable[str(CursorKind.CXX_OVERRIDE_ATTR)] = self.visit_cxx_override_attr
        self._jumpTable[str(CursorKind.ANNOTATE_ATTR)] = self.visit_annotate_attr
        self._jumpTable[str(CursorKind.ASM_LABEL_ATTR)] = self.visit_asm_label_attr
        self._jumpTable[str(CursorKind.PACKED_ATTR)] = self.visit_packed_attr
        self._jumpTable[str(CursorKind.PURE_ATTR)] = self.visit_pure_attr
        self._jumpTable[str(CursorKind.CONST_ATTR)] = self.visit_const_attr
        self._jumpTable[str(CursorKind.NODUPLICATE_ATTR)] = self.visit_noduplicate_attr
        self._jumpTable[str(CursorKind.CUDACONSTANT_ATTR)] = self.visit_cudaconstant_attr
        self._jumpTable[str(CursorKind.CUDADEVICE_ATTR)] = self.visit_cudadevice_attr
        self._jumpTable[str(CursorKind.CUDAGLOBAL_ATTR)] = self.visit_cudaglobal_attr
        self._jumpTable[str(CursorKind.CUDAHOST_ATTR)] = self.visit_cudahost_attr
        self._jumpTable[str(CursorKind.CUDASHARED_ATTR)] = self.visit_cudashared_attr
        self._jumpTable[str(CursorKind.VISIBILITY_ATTR)] = self.visit_visibility_attr
        self._jumpTable[str(CursorKind.DLLEXPORT_ATTR)] = self.visit_dllexport_attr
        self._jumpTable[str(CursorKind.DLLIMPORT_ATTR)] = self.visit_dllimport_attr
        self._jumpTable[str(CursorKind.CONVERGENT_ATTR)] = self.visit_convergent_attr
        self._jumpTable[str(CursorKind.WARN_UNUSED_ATTR)] = self.visit_warn_unused_attr
        self._jumpTable[str(CursorKind.WARN_UNUSED_RESULT_ATTR)] = self.visit_warn_unused_result_attr
        self._jumpTable[str(CursorKind.ALIGNED_ATTR)] = self.visit_aligned_attr
        self._jumpTable[str(CursorKind.PREPROCESSING_DIRECTIVE)] = self.visit_preprocessing_directive
        self._jumpTable[str(CursorKind.MACRO_DEFINITION)] = self.visit_macro_definition
        self._jumpTable[str(CursorKind.MACRO_INSTANTIATION)] = self.visit_macro_instantiation
        self._jumpTable[str(CursorKind.INCLUSION_DIRECTIVE)] = self.visit_inclusion_directive
        self._jumpTable[str(CursorKind.MODULE_IMPORT_DECL)] = self.visit_module_import_decl
        self._jumpTable[str(CursorKind.TYPE_ALIAS_TEMPLATE_DECL)] = self.visit_type_alias_template_decl
        self._jumpTable[str(CursorKind.STATIC_ASSERT)] = self.visit_static_assert
        self._jumpTable[str(CursorKind.FRIEND_DECL)] = self.visit_friend_decl
        self._jumpTable[str(CursorKind.OVERLOAD_CANDIDATE)] = self.visit_overload_candidate

    def visit(self, node: AstNode) -> Generator[Offence, None, None]:
        if node.kind in self._jumpTable:
            return self._jumpTable[node.kind](node)
        else:
            return self.visit_default(node)

    def visit_translation_unit(self, node: AstNode) -> None:
        return

    def visit_unexposed_decl(self, node: AstNode) -> None:
        """A declaration whose specific kind is not exposed via this interface."""
        return

    def visit_struct_decl(self, node: AstNode) -> None:
        """A C or C++ struct."""
        return

    def visit_union_decl(self, node: AstNode) -> None:
        """A C or C++ union."""
        return

    def visit_class_decl(self, node: AstNode) -> None:
        """A C++ class."""
        return

    def visit_enum_decl(self, node: AstNode) -> None:
        """A C or C++ enum."""
        return

    def visit_field_decl(self, node: AstNode) -> None:
        """A field (in C) or non-static data member (in C++) in a struct, union, or C++ class."""
        return

    def visit_enum_constant_decl(self, node: AstNode) -> None:
        """An enumerator constant"""
        return

    def visit_function_decl(self, node: AstNode) -> None:
        """A function"""
        return

    def visit_var_decl(self, node: AstNode) -> None:
        """A variable"""
        return

    def visit_param_decl(self, node: AstNode) -> None:
        """A funtion or method parameter"""
        return

    def visit_objc_interface_decl(self, node: AstNode) -> None:
        """An Objective-C @interface."""
        return

    def visit_typedef_decl(self, node: AstNode) -> None:
        """A typedef"""
        return

    def visit_cxx_method(self, node: AstNode) -> None:
        """A C++ class method"""
        return

    def visit_namespace(self, node: AstNode) -> None:
        """A C++ namespace"""
        return

    def visit_cxx_base_specifier(self, node: AstNode) -> None:
        return

    def visit_linkage_spec(self, node: AstNode) -> None:
        return

    def visit_constructor(self, node: AstNode) -> None:
        return

    def visit_destructor(self, node: AstNode) -> None:
        return

    def visit_conversion_function(self, node: AstNode) -> None:
        return

    def visit_template_type_parameter(self, node: AstNode) -> None:
        return

    def visit_template_non_type_parameter(self, node: AstNode) -> None:
        return

    def visit_template_template_parameter(self, node: AstNode) -> None:
        return

    def visit_function_template(self, node: AstNode) -> None:
        return

    def visit_class_template(self, node: AstNode) -> None:
        return

    def visit_class_template_partial_specialization(self, node: AstNode) -> None:
        return

    def visit_namespace_alias(self, node: AstNode) -> None:
        return

    def visit_using_directive(self, node: AstNode) -> None:
        return

    def visit_using_declaration(self, node: AstNode) -> None:
        return

    def visit_type_alias_decl(self, node: AstNode) -> None:
        return

    def visit_objc_synthesize_decl(self, node: AstNode) -> None:
        return

    def visit_objc_dynamic_decl(self, node: AstNode) -> None:
        return

    def visit_cxx_access_spec_decl(self, node: AstNode) -> None:
        return

    def visit_objc_super_class_ref(self, node: AstNode) -> None:
        return

    def visit_objc_protocol_ref(self, node: AstNode) -> None:
        return

    def visit_objc_class_ref(self, node: AstNode) -> None:
        return

    def visit_type_ref(self, node: AstNode) -> None:
        return

    def visit_template_ref(self, node: AstNode) -> None:
        return

    def visit_namespace_ref(self, node: AstNode) -> None:
        return

    def visit_member_ref(self, node: AstNode) -> None:
        return

    def visit_label_ref(self, node: AstNode) -> None:
        return

    def visit_overloaded_decl_ref(self, node: AstNode) -> None:
        return

    def visit_variable_ref(self, node: AstNode) -> None:
        return

    def visit_invalid_file(self, node: AstNode) -> None:
        return

    def visit_no_decl_found(self, node: AstNode) -> None:
        return

    def visit_not_implemented(self, node: AstNode) -> None:
        return

    def visit_invalid_code(self, node: AstNode) -> None:
        return

    def visit_unexposed_expr(self, node: AstNode) -> None:
        return

    def visit_decl_ref_expr(self, node: AstNode) -> None:
        return

    def visit_member_ref_expr(self, node: AstNode) -> None:
        return

    def visit_call_expr(self, node: AstNode) -> None:
        return

    def visit_objc_message_expr(self, node: AstNode) -> None:
        return

    def visit_block_expr(self, node: AstNode) -> None:
        return

    def visit_integer_literal(self, node: AstNode) -> None:
        return

    def visit_floating_literal(self, node: AstNode) -> None:
        return

    def visit_imaginary_literal(self, node: AstNode) -> None:
        return

    def visit_string_literal(self, node: AstNode) -> None:
        return

    def visit_character_literal(self, node: AstNode) -> None:
        return

    def visit_paren_expr(self, node: AstNode) -> None:
        return

    def visit_unary_operator(self, node: AstNode) -> None:
        return

    def visit_array_subscript_expr(self, node: AstNode) -> None:
        return

    def visit_binary_operator(self, node: AstNode) -> None:
        return

    def visit_compound_assignment_operator(self, node: AstNode) -> None:
        return

    def visit_conditional_operator(self, node: AstNode) -> None:
        return

    def visit_cstyle_cast_expr(self, node: AstNode) -> None:
        return

    def visit_compound_literal_expr(self, node: AstNode) -> None:
        return

    def visit_init_list_expr(self, node: AstNode) -> None:
        return

    def visit_addr_label_expr(self, node: AstNode) -> None:
        return

    def visit_stmtexpr(self, node: AstNode) -> None:
        return

    def visit_generic_selection_expr(self, node: AstNode) -> None:
        return

    def visit_gnu_null_expr(self, node: AstNode) -> None:
        return

    def visit_cxx_static_cast_expr(self, node: AstNode) -> None:
        return

    def visit_cxx_dynamic_cast_expr(self, node: AstNode) -> None:
        return

    def visit_cxx_reinterpret_cast_expr(self, node: AstNode) -> None:
        return

    def visit_cxx_const_cast_expr(self, node: AstNode) -> None:
        return

    def visit_cxx_functional_cast_expr(self, node: AstNode) -> None:
        return

    def visit_cxx_typeid_expr(self, node: AstNode) -> None:
        return

    def visit_cxx_bool_literal_expr(self, node: AstNode) -> None:
        return

    def visit_cxx_null_ptr_literal_expr(self, node: AstNode) -> None:
        return

    def visit_cxx_this_expr(self, node: AstNode) -> None:
        return

    def visit_cxx_throw_expr(self, node: AstNode) -> None:
        return

    def visit_cxx_new_expr(self, node: AstNode) -> None:
        return

    def visit_cxx_delete_expr(self, node: AstNode) -> None:
        return

    def visit_cxx_unary_expr(self, node: AstNode) -> None:
        return

    def visit_objc_string_literal(self, node: AstNode) -> None:
        return

    def visit_objc_encode_expr(self, node: AstNode) -> None:
        return

    def visit_objc_selector_expr(self, node: AstNode) -> None:
        return

    def visit_objc_protocol_expr(self, node: AstNode) -> None:
        return

    def visit_objc_bridge_cast_expr(self, node: AstNode) -> None:
        return

    def visit_pack_expansion_expr(self, node: AstNode) -> None:
        return

    def visit_size_of_pack_expr(self, node: AstNode) -> None:
        return

    def visit_lambda_expr(self, node: AstNode) -> None:
        return

    def visit_obj_bool_literal_expr(self, node: AstNode) -> None:
        return

    def visit_obj_self_expr(self, node: AstNode) -> None:
        return

    def visit_omp_array_section_expr(self, node: AstNode) -> None:
        return

    def visit_objc_availability_check_expr(self, node: AstNode) -> None:
        return

    def visit_unexposed_stmt(self, node: AstNode) -> None:
        return

    def visit_label_stmt(self, node: AstNode) -> None:
        return

    def visit_compound_stmt(self, node: AstNode) -> None:
        return

    def visit_case_stmt(self, node: AstNode) -> None:
        return

    def visit_default_stmt(self, node: AstNode) -> None:
        return

    def visit_if_stmt(self, node: AstNode) -> None:
        return

    def visit_switch_stmt(self, node: AstNode) -> None:
        return

    def visit_while_stmt(self, node: AstNode) -> None:
        return

    def visit_do_stmt(self, node: AstNode) -> None:
        return

    def visit_for_stmt(self, node: AstNode) -> None:
        return

    def visit_goto_stmt(self, node: AstNode) -> None:
        return

    def visit_indirect_goto_stmt(self, node: AstNode) -> None:
        return

    def visit_continue_stmt(self, node: AstNode) -> None:
        return

    def visit_break_stmt(self, node: AstNode) -> None:
        return

    def visit_return_stmt(self, node: AstNode) -> None:
        return

    def visit_asm_stmt(self, node: AstNode) -> None:
        return

    def visit_objc_at_try_stmt(self, node: AstNode) -> None:
        return

    def visit_objc_at_catch_stmt(self, node: AstNode) -> None:
        return

    def visit_objc_at_finally_stmt(self, node: AstNode) -> None:
        return

    def visit_objc_at_throw_stmt(self, node: AstNode) -> None:
        return

    def visit_objc_at_synchronized_stmt(self, node: AstNode) -> None:
        return

    def visit_objc_autorelease_pool_stmt(self, node: AstNode) -> None:
        return

    def visit_objc_for_collection_stmt(self, node: AstNode) -> None:
        return

    def visit_cxx_catch_stmt(self, node: AstNode) -> None:
        return

    def visit_cxx_try_stmt(self, node: AstNode) -> None:
        return

    def visit_cxx_for_range_stmt(self, node: AstNode) -> None:
        return

    def visit_seh_try_stmt(self, node: AstNode) -> None:
        return

    def visit_seh_except_stmt(self, node: AstNode) -> None:
        return

    def visit_seh_finally_stmt(self, node: AstNode) -> None:
        return

    def visit_ms_asm_stmt(self, node: AstNode) -> None:
        return

    def visit_null_stmt(self, node: AstNode) -> None:
        return

    def visit_decl_stmt(self, node: AstNode) -> None:
        return

    def visit_omp_parallel_directive(self, node: AstNode) -> None:
        return

    def visit_omp_simd_directive(self, node: AstNode) -> None:
        return

    def visit_omp_for_directive(self, node: AstNode) -> None:
        return

    def visit_omp_sections_directive(self, node: AstNode) -> None:
        return

    def visit_omp_section_directive(self, node: AstNode) -> None:
        return

    def visit_omp_single_directive(self, node: AstNode) -> None:
        return

    def visit_omp_parallel_for_directive(self, node: AstNode) -> None:
        return

    def visit_omp_parallel_sections_directive(self, node: AstNode) -> None:
        return

    def visit_omp_task_directive(self, node: AstNode) -> None:
        return

    def visit_omp_master_directive(self, node: AstNode) -> None:
        return

    def visit_omp_critical_directive(self, node: AstNode) -> None:
        return

    def visit_omp_taskyield_directive(self, node: AstNode) -> None:
        return

    def visit_omp_barrier_directive(self, node: AstNode) -> None:
        return

    def visit_omp_taskwait_directive(self, node: AstNode) -> None:
        return

    def visit_omp_flush_directive(self, node: AstNode) -> None:
        return

    def visit_seh_leave_stmt(self, node: AstNode) -> None:
        return

    def visit_omp_ordered_directive(self, node: AstNode) -> None:
        return

    def visit_omp_atomic_directive(self, node: AstNode) -> None:
        return

    def visit_omp_for_simd_directive(self, node: AstNode) -> None:
        return

    def visit_omp_parallelforsimd_directive(self, node: AstNode) -> None:
        return

    def visit_omp_target_directive(self, node: AstNode) -> None:
        return

    def visit_omp_teams_directive(self, node: AstNode) -> None:
        return

    def visit_omp_taskgroup_directive(self, node: AstNode) -> None:
        return

    def visit_omp_cancellation_point_directive(self, node: AstNode) -> None:
        return

    def visit_omp_cancel_directive(self, node: AstNode) -> None:
        return

    def visit_omp_target_data_directive(self, node: AstNode) -> None:
        return

    def visit_omp_task_loop_directive(self, node: AstNode) -> None:
        return

    def visit_omp_task_loop_simd_directive(self, node: AstNode) -> None:
        return

    def visit_omp_distribute_directive(self, node: AstNode) -> None:
        return

    def visit_omp_target_enter_data_directive(self, node: AstNode) -> None:
        return

    def visit_omp_target_exit_data_directive(self, node: AstNode) -> None:
        return

    def visit_omp_target_parallel_directive(self, node: AstNode) -> None:
        return

    def visit_omp_target_parallelfor_directive(self, node: AstNode) -> None:
        return

    def visit_omp_target_update_directive(self, node: AstNode) -> None:
        return

    def visit_omp_distribute_parallelfor_directive(self, node: AstNode) -> None:
        return

    def visit_omp_distribute_parallel_for_simd_directive(self, node: AstNode) -> None:
        return

    def visit_omp_distribute_simd_directive(self, node: AstNode) -> None:
        return

    def visit_omp_target_parallel_for_simd_directive(self, node: AstNode) -> None:
        return

    def visit_omp_target_simd_directive(self, node: AstNode) -> None:
        return

    def visit_omp_teams_distribute_directive(self, node: AstNode) -> None:
        return

    def visit_unexposed_attr(self, node: AstNode) -> None:
        return

    def visit_ib_action_attr(self, node: AstNode) -> None:
        return

    def visit_ib_outlet_attr(self, node: AstNode) -> None:
        return

    def visit_ib_outlet_collection_attr(self, node: AstNode) -> None:
        return

    def visit_cxx_final_attr(self, node: AstNode) -> None:
        return

    def visit_cxx_override_attr(self, node: AstNode) -> None:
        return

    def visit_annotate_attr(self, node: AstNode) -> None:
        return

    def visit_asm_label_attr(self, node: AstNode) -> None:
        return

    def visit_packed_attr(self, node: AstNode) -> None:
        return

    def visit_pure_attr(self, node: AstNode) -> None:
        return

    def visit_const_attr(self, node: AstNode) -> None:
        return

    def visit_noduplicate_attr(self, node: AstNode) -> None:
        return

    def visit_cudaconstant_attr(self, node: AstNode) -> None:
        return

    def visit_cudadevice_attr(self, node: AstNode) -> None:
        return

    def visit_cudaglobal_attr(self, node: AstNode) -> None:
        return

    def visit_cudahost_attr(self, node: AstNode) -> None:
        return

    def visit_cudashared_attr(self, node: AstNode) -> None:
        return

    def visit_visibility_attr(self, node: AstNode) -> None:
        return

    def visit_dllexport_attr(self, node: AstNode) -> None:
        return

    def visit_dllimport_attr(self, node: AstNode) -> None:
        return

    def visit_convergent_attr(self, node: AstNode) -> None:
        return

    def visit_warn_unused_attr(self, node: AstNode) -> None:
        return

    def visit_warn_unused_result_attr(self, node: AstNode) -> None:
        return

    def visit_aligned_attr(self, node: AstNode) -> None:
        return

    def visit_preprocessing_directive(self, node: AstNode) -> None:
        return

    def visit_macro_definition(self, node: AstNode) -> None:
        return

    def visit_macro_instantiation(self, node: AstNode) -> None:
        return

    def visit_inclusion_directive(self, node: AstNode) -> None:
        return

    def visit_module_import_decl(self, node: AstNode) -> None:
        return

    def visit_type_alias_template_decl(self, node: AstNode) -> None:
        return

    def visit_static_assert(self, node: AstNode) -> None:
        return

    def visit_friend_decl(self, node: AstNode) -> None:
        return

    def visit_overload_candidate(self, node: AstNode) -> None:
        return

    def visit_default(self, node: AstNode) -> None:
        return

    def final_call(self) -> None:
        """This method will be called after the Tree has been searched."""
        return



from typing import List
from dataclasses import dataclass


@dataclass
class Location():
    line: int
    col: int


@dataclass
class Token():
    kind: str
    value: str
    location_start: Location
    # location_end: Location


class AstNode():
    """Basic data structure containing AST information."""
    def __init__(self, kind: str, filename: str, line: int, col: int, end_line: int,
                 end_col: int, value: str, indet_level: int, val_type: str,
                 parent: "AstNode", tokens: List["Token"]) -> None:

        self.kind: str = kind
        self.location: Location = Location(line, col)
        self.filename = filename
        self.end_location: Location = Location(end_line, end_col)
        self.value: str = value
        self.parent: "AstNode" = parent
        self.children: List["AstNode"] = []
        self.indent_level: int = indet_level
        self.type: str = val_type
        self.tokens: List["Token"] = tokens

    def add_child(self, child: "AstNode") -> None:
        self.children.append(child)

    def __str__(self):
        out = "  " * self.indent_level
        out += str(self.location.line) + ":" + str(self.location.col) + "-" + \
            str(self.end_location.line) + ":" + str(self.end_location.col) + "  " + \
            str(self.kind) + " " + self.value + ": " + self.type
        return out

    def get_older_Siblings(self) -> List["AstNode"]:
        if self.parent is None:
            return []
        return self.parent.children[:self.parent.children.index(self)]

    def get_younger_Siblings(self) -> List["AstNode"]:
        if self.parent is None:
            return []
        return self.parent.children[self.parent.children.index(self):]

    def printTokens(self):
        for t in self.tokens:
            print("  " * self.indent_level + "|" + t.value.replace("\n", "\n|") + f" :{t.kind} {t.location_start.line}:{t.location_start.col}")
